"""
# C++ Converter API

This module handles converting models to C++ format.

Functions:
    array_dump: Generate C array of values from flattened numpy array.
    xxd_c_dump: Generate C array of values from binary file.

"""

from .utils import array_dump, xxd_c_dump
