"""
# :material-chart-ppf: Losses API

The `helia_edge.losses` module provides additional loss functions to train models.

## Available Losses

- **[SimCLR](./simclr)**: A contrastive learning cosine similarity loss function

"""

from . import simclr
